@@ -9,13 +9,21 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-# pylint: disable=R0901, R0904
-from adcm_client.wrappers.api import ADCMApiWrapper
-from adcm_client.util import stream
+# pylint: disable=R0901, R0904, W0401, C0302
+
+import logging
 from contextlib import contextmanager
-from .base import strip_none_keys, BaseAPIObject, BaseAPIListObject, ObjectNotFound, ADCMApiError
+
+from coreapi.exceptions import ErrorMessage
 from version_utils import rpm
 
+from adcm_client.base import (
+    ActionHasIssues, ADCMApiError, BaseAPIListObject, BaseAPIObject, ObjectNotFound,
+    TooManyArguments, strip_none_keys
+)
+from adcm_client.util import stream
+from adcm_client.wrappers.api import ADCMApiWrapper
+
 # If we are running the client from tests with Allure we expected that code
 # to trace steps in Allure UI.
 # But in case of running client outside of testing Allure is useless in virtualenv.
@@ -26,6 +34,10 @@ try:
 except ImportError:
     ALLURE = False
 
+# Init logger
+logger = logging.getLogger()
+logger.setLevel(logging.DEBUG)
+
 
 # That is trick which is allmost the same that in _allure.py::StepContext
 # We have a function that can be used as contextmanager and decorator
@@ -118,6 +130,12 @@ class Bundle(BaseAPIObject):
             raise IncorrectPrototypeType
         return prototype.cluster(**args)
 
+    def license(self):
+        return self._subcall("license", "read")
+
+    def license_accept(self):
+        self._subcall("license", "accept", "update")
+
 
 class BundleList(BaseAPIListObject):
     _ENTRY_CLASS = Bundle
@@ -182,7 +200,6 @@ class ServicePrototype(Prototype):
     shared = None
     display_name = None
     required = None
-    shared = None
     components = None
     exports = None
     imports = None
@@ -439,9 +456,6 @@ class Cluster(_BaseObject):
     def imports(self):
         raise NotImplementedError
 
-    def issue(self):
-        raise NotImplementedError
-
     def upgrade(self, **args) -> "Upgrade":
         return self._subobject(Upgrade, **args)
 
@@ -554,6 +568,7 @@ class Component(BaseAPIObject):
     description = None
     constraint = None
     params = None
+    prototype_id = None
 
     @property
     def service_id(self):
@@ -632,9 +647,18 @@ class Action(BaseAPIObject):
     type = None
     url = None
     subs = None
+    config = None
 
-    def config(self):
-        raise NotImplementedError
+    def _get_config(self):
+        config = dict()
+        for item in self.config['config']:
+            if item['type'] == 'group':
+                config[item['name']] = dict()
+            elif item['subname']:
+                config[item['name']][item['subname']] = item['value']
+            else:
+                config[item['name']] = item['value']
+        return config
 
     def log_files(self):
         raise NotImplementedError
@@ -647,7 +671,32 @@ class Action(BaseAPIObject):
 
     def run(self, **args) -> "Task":
         with allure_step("Run action {}".format(self.name)):
-            data = self._subcall("run", "create", **args)
+
+            if 'config' in args and 'config_diff' in args:
+                raise TypeError("only one argument is expected 'config' or 'config_diff'")
+
+            if 'config' not in args:
+                args['config'] = self._get_config()
+
+                if 'config_diff' in args:
+                    config_diff = args.pop('config_diff')
+                    for item in self.config['config']:
+                        if item['type'] == 'group':
+                            continue
+
+                        key, subkey = item['name'], item['subname']
+                        if subkey and subkey in config_diff.get(key, {}):
+                            args['config'][key][subkey] = config_diff[key][subkey]
+                        elif not subkey and config_diff.get(key):
+                            args['config'][key] = config_diff[key]
+
+            try:
+                data = self._subcall("run", "create", **args)
+            except ErrorMessage as error:
+                if (getattr(error.error, 'title', '') == '409 Conflict'
+                        and 'has issues' in getattr(error.error, '_data', {}).get('desc', '')):
+                    raise ActionHasIssues
+                raise error
             return Task(self._api, task_id=data["id"])
 
 
@@ -687,28 +736,54 @@ class Task(BaseAPIObject):
         return JobList(self._api, paging=paging, path_args=dict(task_id=self.id), **args)
 
     @allure_step("Wait for task end")
-    def wait(self, timeout=None):
-        return self.wait_for_attr("status",
-                                  self._END_STATUSES,
-                                  timeout=timeout)
+    def wait(self, timeout=None, log_failed=True):
+        status = self.wait_for_attr("status",
+                                    self._END_STATUSES,
+                                    timeout=timeout)
+        if log_failed and status == "failed":
+            self._log_jobs('failed')
+        return status
 
     @allure_step("Wait for task to success.")
     def try_wait(self, timeout=None):
         status = self.wait(timeout=timeout)
 
         if status == "failed":
-            for job in self.job_list(status="failed"):
-                for file in job.log_files:
-                    print(self._api.client.get(file["url"])["content"])
             raise TaskFailed
 
         return status
 
+    def _log_jobs(self, status):
+        for job in self.job_list(status=status):
+            for file in job.log_files:
+                response = self._api.client.get(file["url"])
+                if 'content' in response:
+                    logger.error(response["content"])
+
 
 class TaskList(BaseAPIListObject):
     _ENTRY_CLASS = Task
 
 
+##################################################
+#              L O G
+##################################################
+class Log(BaseAPIObject):
+    IDNAME = 'log_id'
+    PATH = ['job', 'log']
+    SUBPATH = ['log']
+    id = None
+    name = None
+    type = None
+    format = None
+    body = None
+
+
+class LogList(BaseAPIListObject):
+    _ENTRY_CLASS = Log
+    SUBPATH = ['log']
+
+
 ##################################################
 #              J O B
 ##################################################
@@ -726,6 +801,7 @@ class Job(BaseAPIObject):
     log_files = None
     task_id = None
 
+    # FIXME: remove method __init__, deal with argument path_args
     def __init__(self, api: ADCMApiWrapper, path=None, path_args=None, **args):
         super().__init__(api, path, **args)
 
@@ -734,6 +810,12 @@ class Job(BaseAPIObject):
                                   self._END_STATUSES,
                                   timeout=timeout)
 
+    def log(self, **kwargs) -> "Log":
+        return self._subobject(Log, **kwargs)
+
+    def log_list(self, paging=None, **kwargs) -> "LogList":
+        return self._subobject(LogList, paging=paging, **kwargs)
+
 
 class JobList(BaseAPIListObject):
     _ENTRY_CLASS = Job
@@ -793,7 +875,7 @@ class ADCM(BaseAPIObject):
 #              C L I E N T
 ##################################################
 class ADCMClient:
-    _MIN_VERSION = "2019.10.01.00"
+    _MIN_VERSION = "2020.04.21.00"
 
     def __init__(self, api=None, url=None, user=None, password=None):
         if api is not None:
@@ -907,8 +989,22 @@ class ADCMClient:
         return self.bundle(bundle_id=data['id'])
 
     @allure_step('Upload bundle from "{1}"')
-    def upload_from_fs(self, dirname) -> Bundle:
-        return self._upload(stream.file(dirname))
+    def upload_from_fs(self, dirname, **args) -> Bundle:
+        streams = stream.file(dirname, **args)
+        if len(streams) > 1:
+            raise TooManyArguments('upload_bundle_from_fs is not capable with building multiple \
+                bundle editions from one dir. Use upload_from_fs_all instead.')
+        return self._upload(streams[0])
+
+    @allure_step('Upload bundles from "{1}"')
+    def upload_from_fs_all(self, dirname, **args) -> BundleList:
+        streams = stream.file(dirname, **args)
+        # Create empty bundle list by filtering on wittingly nonexisting field
+        # and value.
+        result = BundleList(self._api, empty_bundlelist='not_existing_bundle')
+        for st in streams:
+            result.append(self._upload(st))
+        return result
 
     @allure_step('Upload bundle from "{1}"')
     def upload_from_url(self, url) -> Bundle:
